#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import asyncio
import inspect
import itertools
import logging
import os
import shlex
import traceback
from collections import defaultdict
from concurrent import futures
from contextlib import suppress
from datetime import timedelta
from functools import partial
from uuid import UUID

from docopt import DocoptExit, docopt
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.eventloop import use_asyncio_event_loop
from prompt_toolkit.history import FileHistory
from prompt_toolkit.patch_stdout import patch_stdout

from bluetooth_mesh.application import Application, Element
from bluetooth_mesh.apps import get_plugin_manager
from bluetooth_mesh.crypto import DeviceKey, NetworkKey
from bluetooth_mesh.messages.config import GATTNamespaceDescriptor
from bluetooth_mesh.models import (
    ConfigClient,
    ConfigServer,
    DebugClient,
    GatewayConfigClient,
    GenericLevelClient,
    GenericOnOffClient,
    HealthClient,
    LightCTLClient,
    LightLightnessClient,
    NetworkDiagnosticClient,
    NetworkDiagnosticSetupClient,
    SceneClient,
)

plugin_manager = get_plugin_manager()


class MeshCompleter(Completer):
    def __init__(self, application):
        self.application = application
        super().__init__()

    def get_completions(self, document, complete_event):
        if not document.is_cursor_at_the_end:
            return

        command = document.text_before_cursor
        if " " not in command:  # try commands
            for cmd in self.application.commands.values():
                if cmd.CMD.startswith(command):
                    yield Completion(cmd.CMD, start_position=-len(command))

        else:
            words = shlex.split(command)
            if command.endswith(" "):  # wo for difference in split and shlex.split
                words.append("")
            if command.startswith(LsCommand.CMD) or "-g" in words:  # try groups
                for group in self.application.network.groups:
                    group_name = group.name or ""

                    if group_name.startswith(words[-1]):
                        if " " in group_name:
                            group_name = '"{}"'.format(group_name)
                        yield Completion(group_name, start_position=-len(words[-1]))

            else:  # try nodes
                for node in {
                    node.uuid.hex[:4] for node in self.application.network.nodes
                }:
                    if node.startswith(words[-1]):
                        yield Completion(node, start_position=-len(words[-1]))


class Command:
    USAGE = """
    Usage:
        %(cmd)s
    """
    CMD = None

    async def __call__(self, application, arguments):
        raise NotImplementedError()

    def get_usage(self, **kwargs):
        return self.USAGE % dict(cmd=self.CMD, **kwargs)

    def __lt__(self, other):
        return self.CMD < other.CMD


class LsCommand(Command):
    USAGE = """
    Usage:
        %(cmd)s [options] [<groups>...]

    Options:
        -l --long
    """

    CMD = "ls"

    async def __call__(self, application, arguments):
        groups = arguments["<groups>"]

        if not groups and arguments["--long"]:
            groups = [group.name for group in application.network.groups]

        if groups:
            nodes = defaultdict(list)

            for node in application.network.nodes:
                group = application.network.get_node_group(node)

                if not groups or group in groups:
                    nodes[group].append(node)

            for group, nodes in sorted(nodes.items(), key=lambda n: n[0]):
                yield "{}:".format(group)
                for node in sorted(nodes, key=lambda n: n.uuid):
                    if arguments["--long"]:
                        yield "\t{} {:04x} {}".format(
                            node.uuid, node.address, node.name
                        )
                    else:
                        yield "\t{}".format(node.name)
        else:
            for group in sorted(
                group.name for group in application.network.groups if group.name
            ):
                yield "\t{}".format(group)


class ModelCommandMixin:
    ELEMENT = None
    MODEL = None

    def get_model(self, application):
        return application.elements[self.ELEMENT][self.MODEL]


class AttentionCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -t --timeout=ATTENTION    Attention timer [default: 5].
        -g --groups
    """

    ELEMENT = 0
    MODEL = HealthClient
    CMD = "attention"

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        attention = int(arguments["--timeout"])

        tasks = []
        if arguments["<uuid>"]:
            tasks = [
                model.attention(node.address, app_index=0, attention=attention)
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            tasks = [
                model.attention_unack(
                    application.network.get_group_address(name, self.MODEL),
                    app_index=0,
                    attention=attention,
                )
                for name in arguments["<groups>"]
            ]

        return asyncio.gather(*tasks)


class RecallSceneCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>...
        %(cmd)s [options] -g <groups>...

    Options:
        -c --scene=SCENE    Number of scene to recall.
        -t --transition-time=TRANSITION-TIME    Transition time in seconds [default: 0].
        -g --groups
    """
    ELEMENT = 0
    MODEL = SceneClient
    CMD = "recall_scene"

    async def __call__(self, application, arguments):
        model = application.elements[0][SceneClient]

        scene_number = int(arguments["--scene"])
        transition_time = int(arguments["--transition-time"])

        tasks = []
        if arguments["<uuid>"]:
            tasks = [
                model.recall_scene_unack(
                    node.address,
                    app_index=0,
                    scene_number=scene_number,
                    transition_time=transition_time,
                )
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            tasks = [
                model.recall_scene_unack(
                    application.network.get_group_address(name, self.MODEL),
                    app_index=0,
                    scene_number=scene_number,
                    transition_time=transition_time,
                )
                for name in arguments["<groups>"]
            ]

        await asyncio.gather(*tasks)


class NodeSelectionCommandMixin:
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...

    Options:
        -g --groups
    """

    @staticmethod
    def get_addresses(application, arguments):
        uuids = arguments.get("<uuid>", [])
        groups = arguments.get("<groups>", [])
        return [
            node.address
            for node in application.network.nodes
            if node.uuid.hex[:4] in uuids
            or application.network.get_node_group(node) in groups
        ]


class ModelGetCommandMixin(ModelCommandMixin, NodeSelectionCommandMixin):
    PARAMETER = None

    def format(self, data):
        return str(data[self.PARAMETER])

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        get = getattr(model, "get_{}".format(self.PARAMETER))
        results = await get(addresses, 0)

        for address, data in results.items():
            node = application.network.get_node(address=address)
            group = application.network.get_node_group(node)

            param = self.format(data) if data is not None else None
            yield "{} | {}: {}".format(group, node.name, param)


class DebugCommand(ModelGetCommandMixin, Command):
    ELEMENT = 0
    MODEL = DebugClient


class UptimeCommand(DebugCommand):
    CMD = "uptime"
    PARAMETER = "uptime"

    def format(self, data):
        return str(timedelta(seconds=data["uptime"]))


class FaultCommand(DebugCommand):
    CMD = "fault"
    PARAMETER = "last_sw_fault"

    def format(self, data):
        return "{}, {}".format(data["fault"], timedelta(seconds=data["time"]))


class VersionCommand(DebugCommand):
    CMD = "version"
    PARAMETER = "firmware_version"

    def format(self, data):
        return str(data["version"])


class IvIndexCommand(DebugCommand):
    CMD = "ivindex"
    PARAMETER = "ivindex"


class ArapCommand(DebugCommand):
    CMD = "arap"
    PARAMETER = "arap_content"

    def format(self, data):
        return "\n" + "\n".join(
            "\t{:04x}: {:8d} ivi={:1d}".format(k, v["sequence"], v["ivi"])
            for k, v in data["nodes"].items()
        )


class StatsCommand(DebugCommand):
    CMD = "stats"
    PARAMETER = "system_stats"

    def format(self, data):
        return "\n" + "\n".join(
            "\t{:>10s}: {:5d}".format(k, v) for k, v in data["stats"].items()
        )


class AppVersionCommand(DebugCommand):
    CMD = "app_version"
    PARAMETER = "app_version"

    def format(self, data):
        return str(data["version"])


class ConfigCommand(ModelGetCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient


class TtlCommand(ConfigCommand):
    CMD = "ttl"
    PARAMETER = "default_ttl"

    def format(self, data):
        return data["TTL"]


class RelayCommand(ConfigCommand):
    CMD = "relay"
    PARAMETER = "relay"

    def format(self, data):
        return "%s <interval: %sms, count: %s>" % (
            data["relay"],
            data["retransmit"]["interval"],
            data["retransmit"]["count"],
        )


class GatewayConfigurationCommand(
    ModelCommandMixin, NodeSelectionCommandMixin, Command
):
    USAGE = (
        """
    Usage:
        %(cmd)s ("""
        "--mtu=MTU | --mac=MAC | --server=HOST:PORT | --reconnect=INTERVAL | "
        "--dns=IP | --ip=IP | --gateway=IP | --netmask=NETMASK | [--get] | "
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL | "
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP |"
        "--mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP --ip=IP --netmask=NETMASK --gateway=IP"
        """) <uuid>

        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL <uuid>
        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP <uuid>
        %(cmd)s --mtu=MTU --mac=MAC --server=HOST:PORT --reconnect=INTERVAL --dns=IP --ip=IP --netmask=NETMASK --gateway=IP <uuid>
        %(cmd)s [--get=TYPE] <uuid>
    Options:
        --mtu=MTU               Set MTU size
        --mac=MAC               Set MAC address
        --server=SERVER:PORT    Set server's name nad port
        --reconnect=INTERVAL    Set reconnect interval
        --dns=IP                Set DNS IP address
        --ip=IP                 Set IP address
        --gateway=IP            Set gateway UP address
        --netmask=NETMASK       Set netmask
        --get                   Get current configuration
    """
    )

    CMD = "gateway"
    ELEMENT = 0
    MODEL = GatewayConfigClient

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        # strip'--' prefixes from options and drop Nones
        kwargs = {
            k.lstrip("-"): v
            for k, v in arguments.items()
            if k.startswith("--") and v is not None
        }

        kwargs.pop("get", None)

        if len(kwargs) > 1:
            method = partial(model.configuration_set, **self.parse_args(kwargs))
        elif len(kwargs) > 0:
            ((name, value),) = kwargs.items()
            method = partial(getattr(model, f"{name}_set"), **self.parse_args(kwargs))
        else:
            method = getattr(model, f"configuration_get")

        tasks = [
            method(destination=node.address, net_index=0)
            for node in application.network.nodes
            if node.uuid.hex[:4] in arguments["<uuid>"]
        ]

        results = await asyncio.gather(*tasks)
        self.parse_status_payload(results[0]["params"]["payload"])

    @staticmethod
    def parse_args(args):
        parsed = {}
        for key in args:
            if key in ("mtu", "reconnect", "netmask"):
                parsed[key] = int(args[key])
            if key in ("mac", "dns", "ip", "gateway"):
                parsed[key] = args[key]
            if key in "server":
                # convert HOST:PORT into a Tuple
                host, port = args[key].split(":")
                parsed[key] = (host, int(port))

        return parsed

    @staticmethod
    def parse_status_payload(payload):
        server_addr = payload["server_address"]

        if server_addr == "":
            server_addr = "NONE"

        parsed = (
            "revision: {}\n"
            "mac: {}\n"
            "mtu: {}\n"
            "ip: {}/{}\n"
            "gateway: {}\n"
            "dns: {}\n"
            "server: {}:{}\n"
            "reconnect_interval: {}\n"
            "dhcp: {}\n"
            "status_code: {}".format(
                payload["chip_revision_id"],
                payload["mac_address"],
                payload["mtu_size"],
                payload["ip_address"],
                payload["netmask"],
                payload["gateway_ip_address"],
                payload["dns_ip_address"],
                server_addr,
                payload["server_port_number"],
                payload["reconnect_interval"],
                str(payload["flags"]),
                str(payload["status_code"]),
            )
        )
        print(parsed)


class GatewayPacketsCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s [options] <uuid>

    Options:
        --clear     Clear packet's error counters
    """

    CMD = "packets"
    ELEMENT = 0
    MODEL = GatewayConfigClient

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        if arguments["--clear"]:
            method = partial(model.packets_clear)
        else:
            method = partial(model.packets_get)

        tasks = [
            method(node.address, net_index=0)
            for node in application.network.nodes
            if node.uuid.hex[:4] in arguments["<uuid>"]
        ]

        results = await asyncio.gather(*tasks)
        self.parse_status_packets(results[0]["params"]["payload"])

    @staticmethod
    def parse_status_packets(payload):
        parsed = (
            "rx_errors: {}\n"
            "tx_errors: {}\n"
            "bandwidth: {}\n"
            "connection_state: {}\n"
            "link_status: {}\n"
            "last_error: {}".format(
                payload["total_eth_rx_errors"],
                payload["total_eth_tx_errors"],
                payload["bandwidth"],
                str(payload["connection_state"]["conn_state"]),
                str(payload["connection_state"]["link_status"]),
                str(payload["connection_state"]["last_error"]),
            )
        )
        print(parsed)


class AclCommand(Command):
    USAGE = """
    Usage:
        %(cmd)s 
        %(cmd)s --grant <uuid> [--devkey <devkey> --netkey <netkey>]
        %(cmd)s --revoke <uuid>

    Options:
        -g --grant=<uuid>      Grant
        -d --devkey=<devkey>   Devkey
        -n --netkey=<netkey>   Netkey
        -r --revoke=<uuid>     Revoke token
    """
    CMD = "acl"

    def format(self, application):
        for uuid, token in application.token_ring.acl():
            try:
                name = application.network.get_node(uuid=uuid).name
            except KeyError:
                name = None
            yield uuid, name, token

    async def __call__(self, application, arguments):
        if arguments["--grant"]:
            uuid = UUID(arguments["--grant"])
            dev_key = arguments["--devkey"]
            net_key = arguments["--netkey"]

            if not all([dev_key, net_key]):
                try:
                    dev_key = application.network.get_node(uuid=uuid).device_key
                    net_key = application.network.network_keys[0]
                except KeyError:
                    logging.getLogger("acl_grant").warning(
                        "Some key data is missing "
                        "but node could not be found in network configuration to automagically download it."
                    )
                    return
            else:
                dev_key = DeviceKey(bytes.fromhex(dev_key))
                net_key = NetworkKey(bytes.fromhex(net_key))

            await application.acl_grant(uuid=uuid, dev_key=dev_key, net_key=net_key)

        if arguments["--revoke"]:
            await application.acl_revoke(uuid=UUID(hex=arguments["--revoke"]))

        for uuid, name, token in self.format(application):
            print("\t%s (%s): %s" % (uuid, name, token))
        print()


class LightCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
            Usage:
                %(cmd)s <uuid>... [--lightness <light>] [--temperature <temp>]
                %(cmd)s -g <groups>... [--lightness <light>] [--temperature <temp>]

            Options:
                -g --groups
                -l --lightness <light>
                -t --temperature <temp>
            """
    ELEMENT = 0
    CMD = "light"
    PARAMETER = "light_status"
    CTL_ELEMENT = 2

    def format(self, data_light, data_ctl):
        return "lightness {present_lightness}, temperature {present_ctl_temperature}".format(
            present_lightness=data_light.get("present_lightness")
            if data_light
            else None,
            present_ctl_temperature=data_ctl.get("present_ctl_temperature")
            if data_ctl
            else None,
        )

    async def __call__(self, application, arguments):
        addresses = self.get_addresses(application, arguments)

        self.MODEL = LightLightnessClient
        model = self.get_model(application)
        if arguments["--lightness"]:
            results_light = await model.set_lightness(
                nodes=addresses, app_index=0, lightness=int(arguments["--lightness"])
            )
        else:
            results_light = await model.get_lightness(nodes=addresses, app_index=0)

        self.MODEL = LightCTLClient
        model = self.get_model(application)
        if arguments["--temperature"]:
            results_ctl = await model.set_ctl(
                nodes=[addr + self.CTL_ELEMENT for addr in addresses],
                app_index=0,
                ctl_temperature=int(arguments["--temperature"]),
            )
        else:
            results_ctl = await model.get_ctl(
                nodes=[addr + self.CTL_ELEMENT for addr in addresses], app_index=0
            )

        for address in addresses:
            node = application.network.get_node(address=address)
            param = self.format(
                results_light[address], results_ctl[address + self.CTL_ELEMENT]
            )
            print("{}: {}".format(node.name, param))


class NetworkTransmissionCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    USAGE = """
        Usage:
            %(cmd)s <uuid>... [--interval <millis>] [--count <count>]
            %(cmd)s -g <groups>... [--interval <millis>] [--count <count>]

        Options:
            -g --groups
            -s --interval <millis>
            -c --count <count>
        """
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "net_transmission"
    PARAMETER = "net_transmission"

    def format(self, data):
        return "interval={}ms, count={}".format(*data)

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        for address in addresses:
            if arguments["--interval"] and arguments["--count"]:
                results = await model.set_network_transmission(
                    address,
                    net_index=0,
                    interval=int(arguments["--interval"]),
                    count=int(arguments["--count"]),
                )
            else:
                results = await model.get_network_transmission(address, net_index=0)

            node = application.network.get_node(address=address)
            param = self.format(results) if results is not None else None
            print("{}: {}".format(node.name, param))


class CompositionDataCommand(ModelCommandMixin, NodeSelectionCommandMixin, Command):
    ELEMENT = 0
    MODEL = ConfigClient
    CMD = "composition"
    PARAMETER = "composition_data"

    async def __call__(self, application, arguments):
        model = self.get_model(application)
        addresses = self.get_addresses(application, arguments)

        for address in addresses:
            data = await model.get_composition_data([address], net_index=0)
            node = application.network.get_node(address=address)

            composition = data[address]["data"]

            yield "{}: CID {}, PID {}, VID {}, CRPL {}, Features {:016b}".format(
                node.name,
                composition["CID"],
                composition["PID"],
                composition["VID"],
                composition["CRPL"],
                composition["features"],
            )

            for i, ele in enumerate(composition["elements"]):
                yield "\tElement {}, location: {}".format(i, ele["location"])
                yield "\t\t   SIG Models: " + ", ".join(
                    "{:04x}".format(mod["model_id"]) for mod in ele["SIG_models"]
                )
                yield "\t\tVendor Models: " + ", ".join(
                    "{:04x}:{:04x}".format(mod["vendor_id"], mod["model_id"])
                    for mod in ele["vendor_models"]
                )


class GenericOnOffCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s <uuid>...
        %(cmd)s -g <groups>...

    Options:
        -g --groups
    """
    TARGET = None
    MODEL = GenericOnOffClient
    ELEMENT = 0
    PARAMETER = "set_onoff"

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        tasks = []
        if arguments["<uuid>"]:
            command = getattr(model, self.PARAMETER)
            tasks = [
                command(node.address, app_index=0, onoff=self.TARGET)
                for node in application.network.nodes
                if node.uuid.hex[:4] in arguments["<uuid>"]
            ]

        elif arguments["<groups>"]:
            command = getattr(model, "{}_unack".format(self.PARAMETER))
            tasks = [
                command(
                    application.network.get_group_address(name, self.MODEL),
                    app_index=0,
                    onoff=self.TARGET,
                )
                for name in arguments["<groups>"]
            ]

        await asyncio.gather(*tasks)


class GenericOnCommand(GenericOnOffCommand):
    CMD = "on"
    TARGET = True


class GenericOffCommand(GenericOnOffCommand):
    CMD = "off"
    TARGET = False


class MorseCommand(ModelCommandMixin, Command):
    USAGE = """
    Usage:
        %(cmd)s -g <group> <text>

    Options:
        -g --group
    """
    CMD = "morse"
    MODEL = GenericLevelClient
    ELEMENT = 0
    PARAMETER = "set_onoff_unack"

    CODE = {
        "A": ".-",
        "B": "-...",
        "C": "-.-.",
        "D": "-..",
        "E": ".",
        "F": "..-.",
        "G": "--.",
        "H": "....",
        "I": "..",
        "J": ".---",
        "K": "-.-",
        "L": ".-..",
        "M": "--",
        "N": "-.",
        "O": "---",
        "P": ".--.",
        "Q": "--.-",
        "R": ".-.",
        "S": "...",
        "T": "-",
        "U": "..-",
        "V": "...-",
        "W": ".--",
        "X": "-..-",
        "Y": "-.--",
        "Z": "--..",
        "1": ".----",
        "2": "..---",
        "3": "...--",
        "4": "....-",
        "5": ".....",
        "6": "-....",
        "7": "--...",
        "8": "---..",
        "9": "----.",
        "0": "-----",
        ".": ".-.-.-",
        ",": "--..--",
        ":": "---...",
        "?": "..--..",
        "'": ".----.",
        "-": "-....-",
        "/": "-..-.",
        "@": ".--.-.",
        "=": "-...-",
    }

    async def __call__(self, application, arguments):
        model = self.get_model(application)

        destination = application.network.get_group_address(
            name=arguments["<group>"], model=self.MODEL
        )

        send_interval = 0.01
        retransmissions = 6

        set_unack = partial(
            model.set_level_unack,
            destination=destination,
            app_index=0,
            retransmissions=retransmissions,
            send_interval=send_interval,
        )

        on = partial(set_unack, level=32767)
        off = partial(set_unack, level=-24576)

        try:
            # warm up time
            await off()
            await asyncio.sleep(0.5)

            for letter in arguments["<text>"].upper():
                code = self.CODE[letter]
                logging.getLogger("morse").info("%s = %s", letter, code)

                for blink in code:
                    await on(delay=retransmissions * send_interval)

                    delay = (
                        retransmissions * send_interval + 0.1 if blink == "." else 0.5
                    )
                    await off(delay=delay)
                    await asyncio.sleep(delay + 0.1)

                await asyncio.sleep(0.3)
        finally:
            await set_unack(level=-32768)


class HelpCommand(Command):
    CMD = "help"

    async def __call__(self, application, arguments):
        for cmd in sorted(application.commands.values()):
            yield "\t{}".format(cmd.CMD)


class PrimaryElement(Element):
    LOCATION = GATTNamespaceDescriptor.FORTY_SECOND

    MODELS = [
        GenericLevelClient,
        LightLightnessClient,
        LightCTLClient,
        GatewayConfigClient,
        ConfigServer,
        ConfigClient,
        HealthClient,
        DebugClient,
        GenericOnOffClient,
        SceneClient,
        NetworkDiagnosticClient,
        NetworkDiagnosticSetupClient,
    ]


application_mixins = itertools.chain(*get_plugin_manager().hook.application_mixins())


class MeshCommandLine(*application_mixins, Application):
    PATH = "/com/silvair/meshcli/v6"

    COMMANDS = [
        HelpCommand,
        LsCommand,
        AttentionCommand,
        RecallSceneCommand,
        UptimeCommand,
        FaultCommand,
        VersionCommand,
        IvIndexCommand,
        ArapCommand,
        StatsCommand,
        AppVersionCommand,
        LightCommand,
        GenericOnCommand,
        GenericOffCommand,
        CompositionDataCommand,
        MorseCommand,
        NetworkTransmissionCommand,
        TtlCommand,
        RelayCommand,
        GatewayConfigurationCommand,
        GatewayPacketsCommand,
        AclCommand,
    ]

    COMPANY_ID = 0xFEE5
    PRODUCT_ID = 0x42
    VERSION_ID = 1
    CRPL = 0xF00

    ELEMENTS = {0: PrimaryElement}

    def __init__(self, loop: asyncio.AbstractEventLoop, arguments):
        super().__init__(loop)
        self.arguments = arguments
        self.history = FileHistory(os.path.expanduser("~/.meshcli_history"))
        self.completer = MeshCompleter(self)
        self.session = PromptSession(
            history=self.history, completer=self.completer, complete_while_typing=False,
        )
        self.commands = {cmd.CMD: cmd() for cmd in self.COMMANDS}
        self._tid = 0

    async def add_keys(self):
        for index, key in self.subnet_keys:
            await self.add_net_key(index, key)

        for index, bound, key in self.app_keys:
            await self.add_app_key(
                net_key_index=bound, app_key_index=index, app_key=key
            )

        for node in self.network.nodes:
            # don't overwrite my own key
            if node.address in range(self.addr, self.addr + len(self.ELEMENTS)):
                continue

            await self.management_interface.import_remote_node(
                node.address, len(node.elements), DeviceKey(node.device_key.bytes)
            )

        debug_client = self.get_model_instance(element=0, model=DebugClient)
        health_client = self.get_model_instance(element=0, model=HealthClient)

        for index, *_ in self.app_keys:
            await debug_client.bind(index)
            await health_client.bind(index)

    async def run(self, command):
        addr, self.network = await self.get_network()

        async with self:
            await self._run(addr, command)

    async def _run(self, addr, command):
        await self.connect(addr)
        await self.add_keys()
        self.logger.info(
            "Loaded network %s, %d nodes", self.network, len(self.network.nodes)
        )

        while True:
            if command is not None:
                self.logger.info("Running command: %s", command)
                line = command
            else:
                line = await self.session.prompt("{}> ".format(self.uuid), async_=True)

                if not line.strip():
                    continue

            cmd, *argv = shlex.split(line)

            handler = self.commands.get(cmd, None)

            if not handler:
                print("Command not found: {}".format(cmd))
                continue

            usage = handler.get_usage()
            try:
                arguments = docopt(usage, argv, help=False)

                lines = []
                result = handler(self, arguments)

                if inspect.isasyncgen(result):
                    async for line in result:
                        lines.append(line)
                elif inspect.isawaitable(result):
                    result = await result

                    if result is not None:
                        lines.append(str())

                if lines:
                    print("\n".join(lines))
            except Exception as ex:
                print(traceback.format_exc())
            except futures.TimeoutError:
                print("Command timed out: {}".format(cmd))
            except KeyboardInterrupt:
                pass
            except DocoptExit:
                print(usage.strip("\n").rstrip())

            if command:
                break


def main():
    doc = """
        Mesh CLI

        Usage:
            meshcli [options] [<command>]
            meshcli -h | --help | --version

        Options:
            -l --login <login>             User login to platform service (email)
            -p --password <password>       User password to platform service (!unsecured!)
            -n --project <project>         Project name to be loaded from platform

            -d --debug
            -h --help                      Show this help message and exit
            --version                      Show version and exit
    """
    use_asyncio_event_loop()
    arguments = docopt(doc, version="stat_checker 0.5")

    logging.basicConfig(
        format="%(asctime)s %(name)-40s %(levelname)-8s %(filename)15s:%(lineno)3s  %(message)s",
        level=logging.INFO,
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    loop = asyncio.get_event_loop()
    mesh_cli = MeshCommandLine(loop, arguments)

    with suppress(EOFError, KeyboardInterrupt), patch_stdout():
        loop.run_until_complete(mesh_cli.run(arguments.get("<command>")))



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import ecdsa
import pytest

from bluetooth_mesh.mesh import GenericProvisioningPDU
from bluetooth_mesh.provisioning import (
    BearerOpcode,
    GenericProvisioningPDUType,
    LinkCloseReason,
    ProvisioningAlgorithm,
    ProvisioningAuthenticationMethod,
    ProvisioningBearerControl,
    ProvisioningEncryption,
    ProvisioningInputOOBAction,
    ProvisioningOutputOOBAction,
    ProvisioningPDU,
    ProvisioningPDUType,
    ProvisioningPublicKeyType,
    ProvisioningStaticOOBType,
)

valid = [
    # fmt: off
    pytest.param(
        bytes.fromhex('0010'),
        dict(
            type=ProvisioningPDUType.INVITE,
            parameters=dict(
                attention=16
            )
        ),
        id="invite"
    ),
    pytest.param(
        bytes.fromhex('010100010101080009050002'),
        dict(
            type=ProvisioningPDUType.CAPABILITIES,
            parameters=dict(
                num_elements=1,
                algorithms={ProvisioningAlgorithm.FIPS_P256_EC},
                public_key_type={ProvisioningPublicKeyType.OOB},
                static_oob_type={ProvisioningStaticOOBType.AVAILABLE},
                output_oob_size=8,
                output_oob_action={ProvisioningOutputOOBAction.BLINK,
                                   ProvisioningOutputOOBAction.OUTPUT_NUMERIC},
                input_oob_size=5,
                input_oob_action={ProvisioningInputOOBAction.TWIST},
            )
        ),
        id="capabilities",
    ),
    pytest.param(
        bytes.fromhex('020000000000'),
        dict(
            type=ProvisioningPDUType.START,
            parameters=dict(
                algorithm=ProvisioningAlgorithm.FIPS_P256_EC,
                public_key=False,
                authentication_method=ProvisioningAuthenticationMethod.NONE,
                authentication_action=0,
                authentication_size=0,
            )
        ),
        id="start, no authentication",
    ),
    pytest.param(
        bytes.fromhex('020001020304'),
        dict(
            type=ProvisioningPDUType.START,
            parameters=dict(
                algorithm=ProvisioningAlgorithm.FIPS_P256_EC,
                public_key=True,
                authentication_method=ProvisioningAuthenticationMethod.OUTPUT,
                authentication_action=ProvisioningOutputOOBAction.OUTPUT_NUMERIC,
                authentication_size=4,
            )
        ),
        id="start, output numeric",
    ),
    pytest.param(
        bytes.fromhex('05b38a114dfdca1fe153bd2c1e0dc46ac2'),
        dict(
            type=ProvisioningPDUType.CONFIRMATION,
            parameters=dict(
                confirmation=bytes.fromhex('b38a114dfdca1fe153bd2c1e0dc46ac2'),
            )
        ),
        id="confirmation",
    ),
    pytest.param(
        bytes.fromhex('068b19ac31d58b124c946209b5db1021b9'),
        dict(
            type=ProvisioningPDUType.RANDOM,
            parameters=dict(
                random=bytes.fromhex('8b19ac31d58b124c946209b5db1021b9'),
            )
        ),
        id="random",
    ),
    pytest.param(
        bytes.fromhex('07'
                      'd0bd7f4a89a2ff6222af59a90a60ad58acfe3123356f5cec29'
                      '73e0ec50783b10c7'),
        dict(
            type=ProvisioningPDUType.DATA,
            parameters=dict(
                encrypted_provisioning_data=bytes.fromhex('d0bd7f4a89a2ff6222af59a90a60ad58acfe3123356f5cec29'),
                provisioning_data_mic=bytes.fromhex('73e0ec50783b10c7'),
            )
        ),
        id="data",
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,decoded", valid)
def test_build(encoded, decoded):
    result = ProvisioningPDU.build(obj=decoded)
    assert result == encoded


@pytest.mark.parametrize("encoded,decoded", valid)
def test_parse(encoded, decoded):
    result = ProvisioningPDU.parse(data=encoded)
    assert result == decoded


# NOTE: ecdsa.VerifyingKey does not overload __eq__
@pytest.mark.parametrize(
    # fmt: off
    "encoded,decoded",
    [
        pytest.param(
            bytes.fromhex(
                '03'
                '2c31a47b5779809ef44cb5eaaf5c3e43d5f8faad4a8794cb987e9b03745c78dd'
                '919512183898dfbecd52e2408e43871fd021109117bd3ed4eaf8437743715d4f'
            ),
            dict(
                type=ProvisioningPDUType.PUBLIC_KEY,
                parameters=dict(
                    key=ecdsa.VerifyingKey.from_string(
                        bytes.fromhex(
                            '2c31a47b5779809ef44cb5eaaf5c3e43d5f8faad4a8794cb987e9b03745c78dd'
                            '919512183898dfbecd52e2408e43871fd021109117bd3ed4eaf8437743715d4f'
                        ),
                        curve=ecdsa.NIST256p,
                    )
                )
            ),
            id="public key",
        ),
    ]
    # fmt: on
)
def test_parse_key(encoded, decoded):
    result = ProvisioningPDU.parse(data=encoded)
    assert result["type"] == decoded["type"]
    assert result["parameters"]["key"].to_pem() == decoded["parameters"]["key"].to_pem()


def test_confirmation():
    invite = bytes.fromhex("00")
    capabilities = bytes.fromhex("0100010000000000000000")
    start = bytes.fromhex("0000000000")
    provisioner_key = bytes.fromhex(
        "2c31a47b5779809ef44cb5eaaf5c3e43d5f8faad4a8794cb987e9b03745c78dd"
        "919512183898dfbecd52e2408e43871fd021109117bd3ed4eaf8437743715d4f"
    )
    device_key = bytes.fromhex(
        "f465e43ff23d3f1b9dc7dfc04da8758184dbc966204796eccf0d6cf5e16500cc"
        "0201d048bcbbd899eeefc424164e33c201c2b010ca6b4d43a8a155cad8ecb279"
    )
    ecdh_secret = bytes.fromhex(
        "ab85843a2f6d883f62e5684b38e307335fe6e1945ecd19604105c6f23221eb69"
    )
    random = bytes.fromhex("8b19ac31d58b124c946209b5db1021b9")

    auth = None

    salt, key, confirmation = ProvisioningEncryption.confirmation_encrypt(
        ecdh_secret,
        invite + capabilities + start + provisioner_key + device_key,
        random,
        auth,
    )

    assert confirmation == bytes.fromhex("b38a114dfdca1fe153bd2c1e0dc46ac2")


bearer_ctrl = [
    # fmt: off
    pytest.param(
        bytes.fromhex('03 70cf7c9732a345b691494810d2e9cbf4'),
        dict(
            opcode=BearerOpcode.LINK_OPEN,
            parameters=dict(
                device_uuid=bytes.fromhex("70cf7c9732a345b691494810d2e9cbf4")
            ),
            type=GenericProvisioningPDUType.CONTROL
        ),
        id="link open"
    ),
    pytest.param(
        bytes.fromhex('07'),
        dict(
            opcode=BearerOpcode.LINK_ACK,
            parameters=dict(),
            type=GenericProvisioningPDUType.CONTROL
        ),
        id="link ack"
    ),
    pytest.param(
        bytes.fromhex('0b00'),
        dict(
            opcode=BearerOpcode.LINK_CLOSE,
            parameters=dict(
                reason=LinkCloseReason.SUCCESS
            ),
            type=GenericProvisioningPDUType.CONTROL
        ),
        id="link close"
    ),
    # fmt: on
]


@pytest.mark.parametrize("encoded,decoded", bearer_ctrl)
def test_build_provisioning(encoded, decoded):
    result = ProvisioningBearerControl.build(obj=decoded)
    assert result == encoded


@pytest.mark.parametrize("encoded,decoded", bearer_ctrl)
def test_parse_provisioning(encoded, decoded):
    result = ProvisioningBearerControl.parse(data=encoded)
    assert result == decoded


valid = [
    # fmt: off
    pytest.param(
        [bytes.fromhex('03 70cf7c9732a345b691494810d2e9cbf4')],
        dict(
            type=BearerOpcode.LINK_OPEN,
            parameters=dict(
                device_uuid=bytes.fromhex("70cf7c9732a345b691494810d2e9cbf4")
            ),
        ),
        id="link open"
    ),
    pytest.param(
        [bytes.fromhex('00 0002 14 0000')],
        dict(
            type=ProvisioningPDUType.INVITE,
            parameters=dict(
                attention=0
            )
        ),
        id="invite"
    ),
    pytest.param(
        [bytes.fromhex('00 0006 64 020000000000')],
        dict(
            type=ProvisioningPDUType.START,
            parameters=dict(
                algorithm=ProvisioningAlgorithm.FIPS_P256_EC,
                public_key=False,
                authentication_method=ProvisioningAuthenticationMethod.NONE,
                authentication_action=0,
                authentication_size=0,
            )
        ),
        id="start"
    ),
    pytest.param(
        [bytes.fromhex('0400228b 07d0bd7f4a89a2ff6222af59a90a60ad58acfe31'),
         bytes.fromhex('06 23356f5cec2973e0ec50783b10c7')],
        dict(
            type=ProvisioningPDUType.DATA,
            parameters=dict(
                encrypted_provisioning_data=bytes.fromhex('d0bd7f4a89a2ff6222af59a90a60ad58acfe3123356f5cec29'),
                provisioning_data_mic=bytes.fromhex('73e0ec50783b10c7'),
            )
        ),
        id="data",
    ),
    pytest.param(
        [bytes.fromhex('0800411003f465e43ff23d3f1b9dc7dfc04da8758184dbc9'),
         bytes.fromhex('0666204796eccf0d6cf5e16500cc0201d048bcbbd899eeef'),
         bytes.fromhex('0ac424164e33c201c2b010ca6b4d43a8a155cad8ecb279')],
        dict(
            type=ProvisioningPDUType.PUBLIC_KEY,
            parameters=dict(
                key=ecdsa.VerifyingKey.from_string(
                    bytes.fromhex('f465e43ff23d3f1b9dc7dfc04da8758184dbc966204796eccf0d6cf5e16500cc'
                                  '0201d048bcbbd899eeefc424164e33c201c2b010ca6b4d43a8a155cad8ecb279'),
                    curve=ecdsa.NIST256p,
                )
            )
        ),
        id="public key",
    ),
    pytest.param(
        [bytes.fromhex('000011d3068b19ac31d58b124c946209b5db1021b9')],
        dict(
            type=ProvisioningPDUType.RANDOM,
            parameters=dict(
                random=bytes.fromhex('8b19ac31d58b124c946209b5db1021b9')
            )
        ),
        id="random",
    ),
    pytest.param(
        [bytes.fromhex('01')],
        dict(
            type=ProvisioningPDUType.ACK,
            parameters=dict()
        ),
        id="ack",
    )
    # fmt: on
]


@pytest.mark.parametrize("encoded,decoded", valid)
def test_pack_generic(encoded, decoded):
    result = GenericProvisioningPDU.pack(payload=decoded)
    assert result == encoded


@pytest.mark.parametrize("encoded,decoded", valid)
def test_unpack_generic(encoded, decoded):
    result = GenericProvisioningPDU.unpack(segments=encoded)
    assert result == decoded


prov_params = [
    # fmt: off
    pytest.param(
        dict(
            secret=bytes.fromhex('ab85843a2f6d883f62e5684b38e307335fe6e1945ecd19604105c6f23221eb69'),

            confirmation_salt=bytes.fromhex('5faabe187337c71cc6c973369dcaa79a'),
            random_provisioner=bytes.fromhex('8b19ac31d58b124c946209b5db1021b9'),
            random_device=bytes.fromhex('55a2a2bca04cd32ff6f346bd0a0c1a3a'),

            net_key=bytes.fromhex('efb2255e6422d330088e09bb015ed707'),
            net_key_index=b"\x05\x67",
            flags=b"\x00",
            ivindex=b"\x01\x02\x03\x04",
            address=b"\x0b\x0c",

            prov_data=bytes.fromhex('efb2255e6422d330088e09bb015ed707056700010203040b0c'),
            enc_provisioning_data=bytes.fromhex('d0bd7f4a89a2ff6222af59a90a60ad58acfe3123356f5cec29'),
            mic=bytes.fromhex('73e0ec50783b10c7'),
            salt=bytes.fromhex('a21c7d45f201cf9489a2fb57145015b4')
        ),
        id=""
    )
    # fmt: on
]


@pytest.mark.parametrize("params", prov_params)
def test_encrypt_provisioning_data(params):

    prov_data = (
        params["net_key"]
        + params["net_key_index"]
        + params["flags"]
        + params["ivindex"]
        + params["address"]
    )

    assert prov_data == params["prov_data"]

    enc_provisioning_data = ProvisioningEncryption.data_encrypt(
        secret=params["secret"],
        inputs=params["confirmation_salt"]
        + params["random_provisioner"]
        + params["random_device"],
        data=prov_data,
    )

    assert enc_provisioning_data == params["enc_provisioning_data"] + params["mic"]


@pytest.mark.parametrize("params", prov_params)
def test_decrypt_provisioning_data(params):
    enc_provisioning_data = params["enc_provisioning_data"] + params["mic"]

    salt, provisioning_data = ProvisioningEncryption.data_decrypt(
        secret=params["secret"],
        inputs=params["confirmation_salt"]
        + params["random_provisioner"]
        + params["random_device"],
        data=enc_provisioning_data,
    )

    assert provisioning_data == params["prov_data"]
    assert salt == params["salt"]


@pytest.mark.parametrize(
    # fmt :off
    "confirmation_key, confirmation, random, auth",
    [
        pytest.param(
            bytes.fromhex("e31fe046c68ec339c425fc6629f0336f"),
            bytes.fromhex("eeba521c196b52cc2e37aa40329f554e"),
            bytes.fromhex("55a2a2bca04cd32ff6f346bd0a0c1a3a"),
            None,
            id="",
        ),
        pytest.param(
            bytes.fromhex("e31fe046c68ec339c425fc6629f0336f"),
            bytes.fromhex("b38a114dfdca1fe153bd2c1e0dc46ac2"),
            bytes.fromhex("8b19ac31d58b124c946209b5db1021b9"),
            None,
            id="",
        ),
    ]
    # fmt: on
)
def test_confirmation_validate(confirmation_key, confirmation, random, auth):
    assert ProvisioningEncryption.confirmation_validate(
        confirmation_key, confirmation, random, auth
    )


@pytest.mark.parametrize(
    # fmt: off
    "secret, provisioning_salt, device_key",
    [
        pytest.param(
            bytes.fromhex("ab85843a2f6d883f62e5684b38e307335fe6e1945ecd19604105c6f23221eb69"),
            bytes.fromhex("a21c7d45f201cf9489a2fb57145015b4"),
            bytes.fromhex("0520adad5e0142aa3e325087b4ec16d8"),
            id=""
        ),
    ]
    # fmt: on
)
def test_provisioning_device_key(secret, provisioning_salt, device_key):
    assert (
        ProvisioningEncryption.provisioning_device_key(secret, provisioning_salt)
        == device_key
    )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from pytest import fixture, skip

from bluetooth_mesh.crypto import ApplicationKey, DeviceKey, NetworkKey


@fixture
def app_key():
    return ApplicationKey(bytes.fromhex("63964771734fbd76e3b40519d1d94a48"))


@fixture
def net_key():
    return NetworkKey(bytes.fromhex("7dd7364cd842ad18c17c2b820c84c3d6"))


@fixture
def dev_key():
    return DeviceKey(bytes.fromhex("9d6dd0e96eb25dc19a40ed9914f8f03f"))


def test_application_key_id(app_key):
    assert app_key.aid == 0x26


def test_device_key_id(dev_key):
    assert dev_key.aid == 0


def test_encryption_keys_master(net_key):
    nid, encryption_key, privacy_key = net_key.encryption_keys

    assert nid == 0x68
    assert encryption_key == bytes.fromhex("0953fa93e7caac9638f58820220a398e")
    assert privacy_key == bytes.fromhex("8b84eedec100067d670971dd2aa700cf")


def test_network_id(net_key):
    assert net_key.network_id == bytes.fromhex("3ecaff672f673370")


def test_identity_key(net_key):
    assert net_key.identity_key == bytes.fromhex("84396c435ac48560b5965385253e210c")


def test_beacon_key(net_key):
    assert net_key.beacon_key == bytes.fromhex("5423d967da639a99cb02231a83f7d254")



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

from construct import Adapter, BitsInteger, BitStruct, Int8ul, Struct


class TransitionTimeAdapter(Adapter):
    RESOLUTION = {0b00: 0.1, 0b01: 1, 0b10: 10, 0b11: 10 * 60}

    def __init__(self, subcon, allow_unknown=False):
        self.max_steps = 0x3F if allow_unknown else 0x3E
        super().__init__(subcon)

    def _decode(self, obj, context, path):
        return obj["steps"] * self.RESOLUTION[obj["resolution"]]

    def _encode(self, obj, context, path):
        resolution = None
        steps = None
        for range_index, range_value in self.RESOLUTION.items():
            if obj <= range_value * 0x3F:
                resolution = range_index
                steps = obj / range_value
                assert round(steps, 0) <= self.max_steps, "Unknown not allowed"
                break

        assert resolution is not None

        return dict(steps=int(steps), resolution=resolution)


class Delay(Adapter):
    def _decode(self, obj, context, path):
        return obj / 200

    def _encode(self, obj, context, path):
        return int(obj * 200)


# fmt: off
TransitionTime = BitStruct(
    "resolution" / BitsInteger(2),
    "steps" / BitsInteger(6)
)

OptionalSetParameters = Struct(
    "transition_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=False),
    "delay" / Delay(Int8ul)
)
# fmt: on



from construct import Select

from .level import GenericLevelMessage
from .light.ctl import LightCTLMessage, LightCTLSetupMessage
from .light.lightness import LightLightnessMessage, LightLightnessSetupMessage
from .onoff import GenericOnOffMessage

GenericMessage = Select(
    GenericOnOffMessage,
    GenericLevelMessage,
    LightCTLMessage,
    LightCTLSetupMessage,
    LightLightnessMessage,
    LightLightnessSetupMessage,
)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
# pylint: disable=W0223

import enum
import struct

import ecdsa
from construct import (
    Adapter,
    BitsInteger,
    BitStruct,
    Bytes,
    Bytewise,
    Const,
    ExprAdapter,
    GreedyBytes,
    Int8ub,
    Int16ub,
    Int32ub,
    Padding,
    Select,
    Struct,
    Switch,
    this,
)

from bluetooth_mesh.crypto import aes_ccm_decrypt, aes_ccm_encrypt, aes_cmac, k1, s1
from bluetooth_mesh.messages.util import BitList, EmbeddedBitStruct, EnumAdapter


class BearerOpcode(enum.IntEnum):
    LINK_OPEN = 0x00
    LINK_ACK = 0x01
    LINK_CLOSE = 0x02


class ProvisioningPDUType(enum.IntEnum):
    INVITE = 0x00
    CAPABILITIES = 0x01
    START = 0x02
    PUBLIC_KEY = 0x03
    INPUT_COMPLETE = 0x04
    CONFIRMATION = 0x05
    RANDOM = 0x06
    DATA = 0x07
    COMPLETE = 0x08
    FAILED = 0x09
    # ACK type is not in specification and is not actually used in constructing the message
    # it is there just to allow consistent packing/unpacking of GenericProvisioningPDU
    ACK = -1


class GenericProvisioningPDUType(enum.IntEnum):
    START = 0x00
    ACK = 0x01
    CONTINUATION = 0x02
    CONTROL = 0x03


class ProvisioningAlgorithm(enum.IntEnum):
    FIPS_P256_EC = 0


class ProvisioningPublicKeyType(enum.IntEnum):
    OOB = 0


class ProvisioningStaticOOBType(enum.IntEnum):
    AVAILABLE = 0


class ProvisioningOutputOOBAction(enum.IntEnum):
    BLINK = 0
    BEEP = 1
    VIBRATE = 2
    OUTPUT_NUMERIC = 3
    OUTPUT_ALPHANUMERIC = 4


class ProvisioningInputOOBAction(enum.IntEnum):
    PUSH = 0
    TWIST = 1
    INPUT_NUMERIC = 2
    INPUT_ALPHANUMERIC = 3


class ProvisioningAuthenticationMethod(enum.IntEnum):
    NONE = 0
    STATIC = 1
    OUTPUT = 2
    INPUT = 3


class ProvisioningErrorCode(enum.IntEnum):
    PROHIBITED = 0
    INVALID_PDU = 1
    INVALID_FORMAT = 2
    UNEXPECTED_PDU = 3
    CONFIRMATION_FAILED = 4
    OUT_OF_RESOURCES = 5
    DECRYPTION_FAILED = 6
    UNEXPECTED_ERROR = 7
    CANNOT_ASSIGN_ADDRESS = 8


class LinkCloseReason(enum.IntEnum):
    SUCCESS = 0x00
    TIMEOUT = 0x01
    FAIL = 0x02


class ProvisioningPublicKeyAdapter(Adapter):
    def _decode(self, obj, content, path):
        value = obj["x"] + obj["y"]
        return ecdsa.VerifyingKey.from_string(value, curve=ecdsa.NIST256p)

    def _encode(self, obj, content, path):
        value = obj.to_string()
        return dict(x=value[:32], y=value[32:])


# fmt: off
ProvisioningInvite = Struct(
    "attention" / Int8ub,
)

ProvisioningCapabilities = Struct(
    "num_elements" / Int8ub,
    "algorithms" / BitList(2, reversed=True),
    "public_key_type" / BitList(1, reversed=True),
    "static_oob_type" / BitList(1, reversed=True),
    "output_oob_size" / Int8ub,
    "output_oob_action" / BitList(2, reversed=True),
    "input_oob_size" / Int8ub,
    "input_oob_action" / BitList(2, reversed=True),
)

ProvisioningStart = Struct(
    "algorithm" / EnumAdapter(Int8ub, ProvisioningAlgorithm),
    "public_key" / ExprAdapter(Int8ub,
                               lambda obj, ctx: bool(obj),
                               lambda obj, ctx: 1 if obj else 0),
    "authentication_method" / EnumAdapter(Int8ub, ProvisioningAuthenticationMethod),
    "authentication_action" / Switch(
        this.authentication_method,
        {
            ProvisioningAuthenticationMethod.NONE: Const(0, Int8ub),
            ProvisioningAuthenticationMethod.STATIC: Const(0, Int8ub),
            ProvisioningAuthenticationMethod.OUTPUT: EnumAdapter(Int8ub, ProvisioningOutputOOBAction),
            ProvisioningAuthenticationMethod.INPUT: EnumAdapter(Int8ub, ProvisioningInputOOBAction),
        },
    ),
    "authentication_size" / Switch(
        this.authentication_method,
        {
            ProvisioningAuthenticationMethod.NONE: Const(0, Int8ub),
            ProvisioningAuthenticationMethod.STATIC: Const(0, Int8ub),
            ProvisioningAuthenticationMethod.OUTPUT: Int8ub,
            ProvisioningAuthenticationMethod.INPUT: Int8ub,
        },
    ),
)

ProvisioningPublicKey = Struct(
    "key" / ProvisioningPublicKeyAdapter(
        Struct(
            "x" / Bytes(32),
            "y" / Bytes(32),
        ),
    ),
)

ProvisioningInputComplete = Struct()

ProvisioningConfirmation = Struct(
    "confirmation" / Bytes(16),
)

ProvisioningRandom = Struct(
    "random" / Bytes(16),
)

ProvisioningData = Struct(
    "encrypted_provisioning_data" / Bytes(25),
    "provisioning_data_mic" / Bytes(8),
)

ProvisioningPayload = Struct(
    "network_key" / Bytes(16),
    "key_index" / Int16ub,
    "flags" / Int8ub,
    "iv_index" / Int32ub,
    "unicast_address" / Int16ub,
)

ProvisioningComplete = Struct(
)

ProvisioningFailed = Struct(
    "error_code" / EnumAdapter(Int8ub, ProvisioningErrorCode)
)

ProvisioningPDU = Struct(
    *EmbeddedBitStruct(
        "_",
        Padding(2),
        "type" / EnumAdapter(BitsInteger(6), ProvisioningPDUType),
    ),
    "parameters" / Switch(
        this.type,
        {
            ProvisioningPDUType.INVITE: ProvisioningInvite,
            ProvisioningPDUType.CAPABILITIES: ProvisioningCapabilities,
            ProvisioningPDUType.START: ProvisioningStart,
            ProvisioningPDUType.PUBLIC_KEY: ProvisioningPublicKey,
            ProvisioningPDUType.INPUT_COMPLETE: ProvisioningInputComplete,
            ProvisioningPDUType.CONFIRMATION: ProvisioningConfirmation,
            ProvisioningPDUType.RANDOM: ProvisioningRandom,
            ProvisioningPDUType.DATA: ProvisioningData,
            ProvisioningPDUType.COMPLETE: ProvisioningComplete,
            ProvisioningPDUType.FAILED: ProvisioningFailed,
        },
        default=GreedyBytes,
    ),
)

LinkOpen = Struct(
    "device_uuid" / Bytes(16)
)

LinkClose = Struct(
    "reason" / EnumAdapter(Int8ub, LinkCloseReason)
)

ProvisioningBearerControl = BitStruct(
    "opcode" / EnumAdapter(BitsInteger(6), BearerOpcode),
    "type" / Const(GenericProvisioningPDUType.CONTROL, BitsInteger(2)),
    "parameters" / Switch(
        this.opcode,
        {
            BearerOpcode.LINK_OPEN: Struct("device_uuid" / Bytewise(Bytes(16))),
            BearerOpcode.LINK_ACK: Struct(),
            BearerOpcode.LINK_CLOSE: Struct("reason" / EnumAdapter(BitsInteger(8), LinkCloseReason))
        },
        default=GreedyBytes,
    ),
)

TransactionStartPDU = Struct(
    *EmbeddedBitStruct(
        "_",
        "last_segment_number" / BitsInteger(6),
        "type" / Const(GenericProvisioningPDUType.START, BitsInteger(2)),
    ),
    "total_length" / Int16ub,
    "frame_check" / Int8ub,
    "data" / GreedyBytes
)

TransactionContinuationPDU = Struct(
    *EmbeddedBitStruct(
        "_",
        "segment_index" / BitsInteger(6),
        "type" / Const(GenericProvisioningPDUType.CONTINUATION, BitsInteger(2)),
    ),
    "data" / GreedyBytes
)

TransactionPDUSegment = Select(
    TransactionStartPDU,
    TransactionContinuationPDU,
    ProvisioningBearerControl
)

ProvisioningMessage = Struct(
    "link_id" / Bytes(4),
    "transaction_id" / Int8ub,
    "data" / GreedyBytes
)
# fmt: on


class ProvisioningEncryption:
    @staticmethod
    def data_encrypt(secret, inputs, data):
        """ inputs = confirmation_salt + provisioner_random + device_random """
        provisioning_salt = s1(inputs)
        provisioning_key = k1(secret, provisioning_salt, b"prsk")
        provisioning_nonce = k1(secret, provisioning_salt, b"prsn")[-13:]

        return aes_ccm_encrypt(provisioning_key, provisioning_nonce, data, tag_length=8)

    @staticmethod
    def data_decrypt(secret, inputs, data, mic=b""):
        """ inputs = confirmation_salt + provisioner_random + device_random """
        provisioning_salt = s1(inputs)
        provisioning_key = k1(secret, provisioning_salt, b"prsk")
        provisioning_nonce = k1(secret, provisioning_salt, b"prsn")[-13:]

        return (
            provisioning_salt,
            aes_ccm_decrypt(
                provisioning_key, provisioning_nonce, data + mic, tag_length=8
            ),
        )

    @staticmethod
    def provisioning_device_key(secret, provisioning_salt):
        return k1(secret, provisioning_salt, b"prdk")

    @staticmethod
    def confirmation_encrypt(secret, inputs, random, auth=None):
        """ inputs = invite(attention) + capabilities(without opcode) + start(msg) + provisioner_key + device_key """
        confirmation_salt = s1(inputs)
        confirmation_key = k1(secret, confirmation_salt, b"prck")

        return (
            confirmation_salt,
            confirmation_key,
            aes_cmac(confirmation_key, random + struct.pack("16s", auth or b"")),
        )

    @staticmethod
    def confirmation_validate(confirmation_key, confirmation, random, auth=None):
        return confirmation == aes_cmac(
            confirmation_key, random + struct.pack("16s", auth or b"")
        )



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from enum import IntEnum

from construct import (
    Array,
    ExprValidator,
    Int8ul,
    Int16ul,
    Select,
    Struct,
    Switch,
    obj_,
    this,
)

from bluetooth_mesh.messages.generics import (
    Delay,
    TransitionTime,
    TransitionTimeAdapter,
)
from bluetooth_mesh.messages.util import EnumAdapter, Opcode


class SceneOpcode(IntEnum):
    SCENE_GET = 0x8241
    SCENE_RECALL = 0x8242
    SCENE_RECALL_UNACKNOWLEDGED = 0x8243
    SCENE_STATUS = 0x5E
    SCENE_REGISTER_GET = 0x8244
    SCENE_REGISTER_STATUS = 0x8245
    SCENE_STORE = 0x8246
    SCENE_STORE_UNACKNOWLEDGED = 0x8247
    SCENE_DELETE = 0x829E
    SCENE_DELETE_UNACKNOWLEDGED = 0x829F


class SceneStatusCode(IntEnum):
    SUCCESS = 0x00
    SCENE_REGISTER_FULL = 0x01
    SCENE_NOT_FOUND = 0x02


# fmt: off
SceneGet = Struct()

SceneRecallMinimal = Struct(
    "scene_number" / ExprValidator(Int16ul, obj_ > 0),
    "tid" / Int8ul,
)

SceneRecallWithTransition = Struct(
    "scene_number" / ExprValidator(Int16ul, obj_ > 0),
    "tid" / Int8ul,
    "transition_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=False),
    "delay" / Delay(Int8ul),
)

SceneRecall = Select(
    SceneRecallWithTransition,
    SceneRecallMinimal
)

SceneStatusMinimal = Struct(
    "status_code" / EnumAdapter(Int8ul, SceneStatusCode),
    "current_scene" / Int16ul
)

SceneStatusWithTargetScene = Struct(
    "status_code" / EnumAdapter(Int8ul, SceneStatusCode),
    "current_scene" / Int16ul,
    "target_scene" / Int16ul,
    "remaining_time" / TransitionTimeAdapter(TransitionTime, allow_unknown=False),
)

SceneStatus = Select(
    SceneStatusWithTargetScene,
    SceneStatusMinimal
)

SceneRegisterGet = Struct()

SceneRegisterStatus = Struct(
    "status_code" / EnumAdapter(Int8ul, SceneStatusCode),
    "current_scene" / Int16ul,
    "scenes" / Array(16, Int16ul)
)

SceneSetupWithValidation = Struct(
    "scene_number" / ExprValidator(Int16ul, obj_ > 0),
)

SceneSetup = Struct(
    "scene_number" / Int16ul
)

SceneMessage = Struct(
    "opcode" / Opcode(SceneOpcode),
    "params" / Switch(
        this.opcode,
        {
            SceneOpcode.SCENE_GET: SceneGet,
            SceneOpcode.SCENE_RECALL: SceneRecall,
            SceneOpcode.SCENE_RECALL_UNACKNOWLEDGED: SceneRecall,
            SceneOpcode.SCENE_STATUS: SceneStatus,
            SceneOpcode.SCENE_REGISTER_GET: SceneRegisterGet,
            SceneOpcode.SCENE_REGISTER_STATUS: SceneRegisterStatus,
            SceneOpcode.SCENE_STORE: SceneSetupWithValidation,
            SceneOpcode.SCENE_STORE_UNACKNOWLEDGED: SceneSetupWithValidation,
            SceneOpcode.SCENE_DELETE: SceneSetup,
            SceneOpcode.SCENE_DELETE_UNACKNOWLEDGED: SceneSetup,

        }
    )
)
# fmt: on



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import os
from json.decoder import JSONDecodeError

from marshmallow import Schema, ValidationError, fields


class StoredNodeSchema(Schema):
    token = fields.Integer()
    acl = fields.Dict(keys=fields.UUID(), values=fields.Integer())
    network = fields.Dict(keys=fields.String(), values=fields.String())


class TokenRing:
    PATH = "~/.cache/bluetooth-mesh"
    LEGACY_PATH = "~/.cache/bluetooth-mesh-example"

    @property
    def path(self):
        return os.path.expanduser(self.PATH)

    def __init__(self, uuid):
        self.uuid = str(uuid)
        self.schema = StoredNodeSchema()
        self.data = {}

        os.makedirs(self.path, exist_ok=True)
        for path in [self.path, os.path.expanduser(self.LEGACY_PATH)]:
            try:
                with open(os.path.join(path, self.uuid), "r") as tokenfile:
                    r = tokenfile.read()
                    try:
                        self.data = self.schema.loads(r)
                        return
                    except (JSONDecodeError, ValidationError, EOFError):
                        self.data = dict(token=int(r, 16), acl={}, network={})

            except FileNotFoundError:
                self.data = dict(token=0, acl={}, network={})

    def _save(self):
        with open(os.path.join(self.path, self.uuid), "w") as tokenfile:
            tokenfile.write(self.schema.dumps(self.data))

    @property
    def token(self):
        return self.data["token"]

    @token.setter
    def token(self, value):
        self.data["token"] = value
        self._save()

    def acl(self, uuid=None, token=None):
        if all((uuid, token)):
            self.data["acl"][uuid] = token
            self._save()
            return

        return self.data["acl"].get(uuid) if uuid else self.data["acl"].items()

    def drop_acl(self, uuid):
        del self.data["acl"][uuid]
        self._save()



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
from setuptools import find_packages, setup

with open("README.rst", "r") as f:
    long_description = f.read()

# fmt: off
setup(
    name='bluetooth-mesh',
    version='0.2.18',
    author_email='michal.lowas-rzechonek@silvair.com',
    description=(
        'Bluetooth mesh for Python'
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='http://github.com/silvairgit/python-bluetooth-mesh',
    packages=find_packages(exclude=('test*', )),
    python_requires='>=3.6.0,<3.8.0',
    setup_requires=[
        'pytest-runner==4.2',
    ],
    install_requires=[
        'bitstring>=3.1.5',
        'construct==2.9.45',
        'cryptography>=2.3.1',
        'crc==0.3.0',
        'dbus-next==0.1.3',
        'ecdsa==0.15',
        'pluggy>=0.13.1',
        'marshmallow>=3.0.1,<4.0',
    ],
    tests_require=[
        'asynctest==0.12.3',
        'coveralls==2.0.0',
        'pytest==5.0.0',
        'pytest-asyncio==0.10.0',
        'pytest-cov>=2.8.1',
    ],
    extras_require=dict(
        docs=[
            'sphinx>=2.1.2',
            'sphinx-autodoc-typehints>=1.7.0',
        ],
        tools=[
            'prompt-toolkit==2.0.10',
        ],
    ),
    entry_points=dict(
        console_scripts=[
            'meshcli = bluetooth_mesh.apps.meshcli:main [tools]',
        ],
        bluetooth_mesh=[
            'bluetooth_mesh = bluetooth_mesh.apps.mixins',
        ]
    ),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Framework :: AsyncIO',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 (GPLv2)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: System :: Networking',
    ],
)
# fmt: on



"""
This package provides message definitions for vendor-specific models supported
by `Silvair Lighting Firmware`_.

.. _`Silvair Lighting Firmware`: https://silvair.com/solutions/firmware/

"""
# pylint: disable=W0611

from .debug import DebugMessage, DebugSubOpcode
from .gateway_config_server import GatewayConfigMessage, GatewayConfigServerSubOpcode
from .network_diagnostic_server import (
    NetworkDiagnosticServerMessage,
    NetworkDiagnosticServerSubOpcode,
    NetworkDiagnosticSetupServerMessage,
    NetworkDiagnosticSetupServerSubOpcode,
)



#
# python-bluetooth-mesh - Bluetooth Mesh for Python
#
# Copyright (C) 2019  SILVAIR sp. z o.o.
#
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#
import pytest

from bluetooth_mesh.utils import construct_match


@pytest.mark.parametrize(
    "received, expected",
    [
        pytest.param(42, 42,),
        pytest.param([1, 2, 3], [1, 2, 3],),
        pytest.param([1, 2, 3, 4], [1, 2, ..., 4],),
        pytest.param(dict(a=42, b=[1, 2, 3]), dict(a=42, b=[1, 2, 3]),),
        pytest.param(dict(a=42, b=[1, 2, 3]), dict(b=[1, 2, 3]),),
    ],
)
def test_construct_match(received, expected):
    assert construct_match(received, expected)


@pytest.mark.parametrize(
    "received, expected",
    [
        pytest.param(42, 41,),
        pytest.param([1, 2, 3], [1, 2, 3, 4],),
        pytest.param([1, 2, 3, 4], [1, 2, ...],),
        pytest.param(dict(a=42, b=[1, 2, 3]), dict(a=42, b=[1, 3]),),
    ],
)
def test_construct_doesnt_match(received, expected):
    assert not construct_match(received, expected)
